<?php namespace awp;

class DBManager {
	
	private static $instance = null;
	
	private function __construct() 
	{
		//Default filters will only be added once in singleton.
		add_filter("DataTypesValidators", function($table) {  return $this->DefaultTypeValidators(); } );
		add_filter("ValidateFormStructure", function($table) {  return $this->DefaultStructureVlaidator($table); } );
	}
	
	
	public static function getInstance()
	{
		if(!self::$instance)
		{
			self::$instance = new DBManager();
		}

		return self::$instance;
	}
	
	
	private function DefaultTypeValidators($existingValidators = array())
	{
		//why 3,
		//https://stackoverflow.com/questions/10081500/mysql-char-varchar-character-sets-storage-sizes#:~:text=For%20ASCII%2C%20that's%201%20byte,and%20VARCHAR%20are%20the%20same.

		$charBytes = str_contains( strtolower(DB_CHARSET), "utf8") ? 3 : 1;/*askii*/;
		if(!defined("DB_CHAR_BYTES"))
		define("DB_CHAR_BYTES", $charBytes);

		//type descriptions
		//https://www.w3schools.com/mysql/mysql_datatypes.asp#:~:text=In%20MySQL%20there%20are%20three,numeric%2C%20and%20date%20and%20time.

		$newValidators = (object)[
			"char" => function( $fieldData, $input ){ 
				return $this->LengthInputCheck($fieldData, $input) && $this->FixedLengthCheck(255, $input); 
			},
			"varchar" => function( $fieldData, $input ){ 
				return $this->LengthInputCheck($fieldData, $input) && $this->FixedLengthCheck(65535, $input); 
			},
			"binary" => function( $fieldData, $input ){
				return $this->LengthInputCheck($fieldData, $input) && $this->FixedLengthCheck(255/DB_CHAR_BYTES, $input); 
			},
			"varbinary" => function( $fieldData, $input ){ 
				return $this->LengthInputCheck($fieldData, $input) && $this->FixedLengthCheck(65535/DB_CHAR_BYTES, $input); 
			},
			"tinyblob" => function( $fieldData, $input ){ 
				return $this->FixedLengthCheck(255/DB_CHAR_BYTES, $input); 
			},
			"tinytext" => function( $fieldData, $input ){ 
				return $this->FixedLengthCheck(DB_CHAR_BYTES, $input); 
			},
			"text" => function( $fieldData, $input ){ 
				return (isset($fieldData->length) ? $this->LengthInputCheck($fieldData, $input) : true) && $this->FixedLengthCheck(65535/DB_CHAR_BYTES, $input); 
			},
			"blob" => function( $fieldData, $input ){ 
				return $this->LengthInputCheck($fieldData, $input) && $this->FixedLengthCheck(65535/DB_CHAR_BYTES, $input); 
			},
			"enum" => function( $fieldData, $input ){ 
				return $this->EnumInputCheck($fieldData, $input); 
			},
			"set" => function( $fieldData, $input ){ 
				return $this->SetInputCheck($fieldData, $input) && $this->FixedLengthCheck(65535/DB_CHAR_BYTES, $input); 
			},
			"bit" => function( $fieldData, $input ){ 
				return $this->FixedIntInputCheck(64, $input); 
			},
			"tinyint" => function( $fieldData, $input ){ 
				return $this->FixedIntInputCheck(255, $input); 
			},
			"bool" => function( $fieldData, $input ){ 
				return is_bool($input); 
			},
			"boolean" => function( $fieldData, $input ){ 
				return is_bool($input); 
			},
			"smallint" => function( $fieldData, $input ){ 
				return $this->FixedIntInputCheck(65535, $input); 
			},
			"int" => function( $fieldData, $input ){ 
				return $this->FixedIntInputCheck(4294967295, $input); 
			},
			"integer" => function( $fieldData, $input ){ 
				return $this->FixedIntInputCheck(4294967295, $input); 
			},
			"float" => function( $fieldData, $input ){ 
				return is_float($input); 
			},
			"double" => function( $fieldData, $input ){ 
				return is_double($input); 
			},
			"decimal" => function( $fieldData, $input ){ 
				return is_float($input); 
			},
			"date" => function( $fieldData, $input ){ 
				return preg_match("/[1-9][0-9]{3}\-(?:0[1-9]|1[0-2])\-(?:0[1-9]|[12][0-9]|3[01])/", $input); 
			},
			"datetime" => function( $fieldData, $input ){ 
				return preg_match("/[1-9][0-9]{3}\-(?:0[1-9]|1[0-2])\-(?:0[1-9]|[12][0-9]|3[01]) (?:[01][0-9]|2[0-3])(?:\:[0-5][0-9]){2}/", $input); 
			},
			"timestamp" => function( $fieldData, $input ){ 
				return true;/*to complex*/ 
			},
			"time" => function( $fieldData, $input ){ 
				return preg_match("/\-?(?:(?:8[0-3][0-8])|[0-7]?\d?\d)(?:\:[0-5][0-9]){2}/", $input); 
			},
			"year" => function( $fieldData, $input ){ 
				return preg_match("/0{4}|190[1-9]|19[1-9][0-9]|20[0-9]{2}|21[0-5]{2}/", $input); 
			},
		];

		return (object) array_merge((array)$existingValidators, (array)$newValidators);
	}
	
	

	private function FixedIntInputCheck($size, $input)
	{
		if(is_int($input) &&  $input <= $size)
			return true;
	}

	private function EnumInputCheck($fieldData, $input)
	{
		if(is_string($fieldData))
			$fieldData = $this->GetField($fieldData);

		if(is_array($fieldData->length) && in_array($input, $fieldData->length))
			return true;

	}

	private function SetInputCheck($fieldData, $input)
	{
		if(is_string($fieldData))
			$fieldData = $this->GetField($fieldData);

		if(is_string($input))
			$input = explode(",", $input);

		if(!is_array($input) || !is_array($fieldData->length) || count($input) > 64)
			return false;

		for($i=0; $i<count($input); $i++)
		{
			if(!in_array($input[$i], fieldData->length))
			{
				return false;
			}
		}

		return true;
	}

	private function LengthInputCheck($fieldData, $input)
	{

		if(is_string($fieldData))
			$fieldData = $this->GetField($fieldData);

		if(is_string($input) && strlen($input) <= $fieldData->length)
			return true;
	}

	private function FixedLengthCheck($length, $input)
	{
		if(is_string($input) && strlen($input) <= $length)
			return true;
	}
	
	
	
	/**
	 * 
	 * Defalt Structure Validator checks if the field values are formatted correctly when creating your table e.g:
	 * name: no spaces caps or numbers must use _ to space words out.
	 * type: has been registerd in dataTypesValidators.
	 * 
	 * 
	**/
	
	private function DefaultStructureVlaidator($table)
	{
		for($i=0; $i<count($table->fields); $i++)
		{
			$fieldData = $table->fields[$i];

			if(!property_exists($fieldData, "name") || !property_exists($fieldData, "type") )
			{
				echo "Field Data Invalid: must contain name and type properties.";
				var_dump($fieldData);
				die();
			}

			$fieldKeys = array_keys(get_object_vars($fieldData));
			for($j=0; $j<count($fieldKeys); $j++)
			{
				$property = $fieldKeys[$j];
				$value = $fieldData->$property;


				switch($property){
					case "name":
						$this->ValidateNameStructure($table, $value);
						break;
					case "type":
						$this->ValidateTypeStructure($table, $fieldData->name, strtolower($value));
						break;
					case "length":
						//$this->ValidteLenghtStructure($table, $fieldData->name, $value);
						//ahh can be array to
						break;
					case "nullable":
					case "unique":
					case "primary":
					case "primaryKey":
					case "unsigned":
						$this->ValidateBoolStructure($table, $fieldData->name, $value);
						break;
					case "default":
						$this->ValidateDefaultStructure($table, $fieldData->name, $fieldData, $value);
						break;
				}


			}

		}

		return $table->fields;
	}
	

	private function ValidateNameStructure($table, $value)
	{
		if($value != sanitize_key($value))
		{
			echo "Field name: {$value} in table: {$table->name} is invalid.<br /> you can try this: " . sanitize_key($value) . " or " . \SnakeCase($value) . "instead.";
			die();
		}

		return true;
	}

	public function ValidateTypeStructure($table, $field, $value)
	{

		if(!property_exists($table->dataTypesValidators, $value))
		{
			echo "No validation methods for field type: {$value}. for table: {$table->name}, Filed: {$field}<br/> Available Validators: ";
			var_dump(array_keys(get_object_vars($table->dataTypesValidators)));
			die();
		}

		return true;
	}

	private function ValidateLengthStructure($table, $field, $value)
	{
		if(intval($value) <= 0)
		{
			echo "Field Length on Table {$table->name}, Filed: {$field}, must be an int more than 0,";
			die();
		}

		return true;
	}

	private function ValidateBoolStructure($table, $field, $value, $errorMessage = null)
	{
		if(! is_bool($value))
		{
			echo $errorMessage || "value must be a bool on Table {$table->name}, Filed: {$field}";
			die();
		}

		return true;
	}

	private function ValidateDefaultStructure($table, $field, $fieldData, $value)
	{


		if( isset($fieldData->ignoreDefaultCheck) && $fieldData->ignoreDefaultCheck)
			return true;

		$type = $fieldData->type;
		$callable = $table->dataTypesValidators->$type;

		if(!$callable($fieldData, $value))
		{
			echo "Default value failed validation check on Table {$table->name}, Filed: {$field}";
			die();
		}
		
		return true;
	}
	
}


?>